/*
* game.c
*
* Created: 5/22/2018 5:22:21 PM
* Author: Vlad George-Alexandru
*/

/****************************************************************************
*
* Proiectare cu Microprocesoare, 2018
* Author: Vlad George-Alexandru <george.vlad0108@gmail.com>
*
****************************************************************************/

#include <stdio.h>
#include <avr/io.h>
#include <util/delay.h>
#include <inttypes.h>
#include <string.h>

#include "game.h"
#include "vector.h"
#include "ST7735R_TFT.h"

/****************************************************************************
* Preprocessor Definitions
****************************************************************************/

#define GAME_WINDOW_WIDTH  130
#define GAME_WINDOW_HEIGHT 142
#define SCORE_OFFSET       20

#define MORSE_LUNG		1000
#define MORSE_SCURT		100

#define SNAKE_MAX_SEGMENTS       40
#define SNAKE_INIT_SEGMENTS      6
#define SEGMENT_WIDTH            6
#define SEGMENT_HEIGHT           6
#define SEGMENT_SPACE            2
#define SNAKE_SPEED				 1

#define SNAKE_HEAD_COLOR_R		 0x00
#define SNAKE_HEAD_COLOR_G		 0x66
#define SNAKE_HEAD_COLOR_B		 0x00

#define SNAKE_BODY_COLOR_R		 0x00
#define SNAKE_BODY_COLOR_G		 0x66
#define SNAKE_BODY_COLOR_B		 0x00

#define SNAKE_POS_X			     (GAME_WINDOW_WIDTH / 2)
#define SNAKE_POS_Y			     (GAME_WINDOW_HEIGHT / 2)

#define FOOD_NUMBER				 2
#define FOOD_RADIUS				 4

#define FOOD_BLUE_COLOR_R		 0x00
#define FOOD_BLUE_COLOR_G		 0x00
#define FOOD_BLUE_COLOR_B		 0xFF

#define FOOD_RED_COLOR_R		 0xFF
#define FOOD_RED_COLOR_G		 0x00
#define FOOD_RED_COLOR_B		 0x00

#define GAME_WINDOW_BG_R		 0
#define GAME_WINDOW_BG_G		 0
#define GAME_WINDOW_BG_B		 0

/****************************************************************************
* Private Types
****************************************************************************/

enum food_type_e
{
	BLUE,
	RED
};

enum game_status_e
{
	GAME_INIT,
	GAME_RUNNING,
	GAME_LOST,
	GAME_WON,
};

enum food_status_e
{
	FOOD_EATEN,
	FOOD_FRESH
};

enum snake_direction_e
{
	UP,
	DOWN,
	LEFT,
	RIGHT
};

typedef struct color_t
{
	uint8_t r;
	uint8_t g;
	uint8_t b;
} color_t;

typedef struct food_t
{
	/* Food's color. */

	color_t color;

	/* Food's position. */

	vector2i_t pos;

	/* Food's type. */

	uint8_t type;
	
	/* Status */
	
	uint8_t status;
} food_t;

typedef struct segment_t
{
	/* Snake's segments color*/
	
	color_t color;
	
	/* Segment position */
	
	vector2i_t pos;
	
	/* Body segment or head */
	
	uint8_t    head;
} segment_t;

typedef struct snake_t
{
	/* Snake's segments. */
	
	segment_t   segments[SNAKE_MAX_SEGMENTS];
	
	/* Snake's tail segment. */
	
	segment_t   tail;
	
	/* Snake length. */
	
	uint8_t     length;
	
	/* Snake's direction. */
	
	uint8_t     direction;
	
	/* Boost */

	int boost;
} snake_t;

typedef struct game_t
{

	/* Current game status. */

	uint8_t     status;
	
	/* Snake */

	snake_t     snake;
	
	/* Food */
	
	food_t food[FOOD_NUMBER];
	
	/* Score */
	
	int score;
} game_t;

/****************************************************************************
* Private Data
****************************************************************************/

static game_t game;
static int rand_state = 2;
static int current_food_items = 0;
static int buzz_delay = 0;
static int win_sound[] = {4, 8, 8, 4, 4, 4, 4, 4};

/****************************************************************************
* Private Functions Prototypes
****************************************************************************/

static void GAME_draw_snake(void);
static void GAME_draw_segment(int x, int y, uint8_t r, uint8_t g, uint8_t b);
static void GAME_generate_food(int index);
static void GAME_draw_food(int index);
static void GAME_draw_all_food(void);
static void GAME_init_food(void);
static void GAME_physics(void);
static void GAME_draw_score(void);
static void GAME_init_segment(uint8_t head, int pos_x, int pos_y, int index);
static int  GAME_check_food_collision(void);
static int  GAME_check_snake_collsion(void);
static void GAME_buzzer_morse(int tip_morse);
static void GAME_make_sound(int tip_morse, int type);

/****************************************************************************
* Public Functions
****************************************************************************/

void GAME_init(void)
{
	/* Empty screen. */

	ST7735R_FillRect(0, 0, ST7735R_WIDTH - 1, ST7735R_HEIGHT - 1,
	GAME_WINDOW_BG_R, GAME_WINDOW_BG_G,
	GAME_WINDOW_BG_B);

	/* Initial snake length and direction. */
	game.snake.direction = RIGHT;
	game.snake.length = SNAKE_INIT_SEGMENTS;
	game.snake.boost = 0;
	game.status = GAME_INIT;
	game.score = 0;

	/* Init snake's segments. */
	for (int i = 0; i < SNAKE_INIT_SEGMENTS; i++) {
		/* Snake head */
		if (i == 0) {
			GAME_init_segment(1, SNAKE_POS_X, SNAKE_POS_Y, i);
		} else {
			GAME_init_segment(0, SNAKE_POS_X - i * SEGMENT_WIDTH - i * SEGMENT_SPACE, SNAKE_POS_Y, i);
		}
	}
	
	/* Generate food on the map. */
	GAME_init_food();
	
	/* Draw a line to split the score and game area */
	ST7735R_Line(0, SCORE_OFFSET - 2, ST7735R_WIDTH,  SCORE_OFFSET - 2, 0xFF, 0xFF, 0xFF);

	GAME_draw_snake();
	GAME_draw_all_food();

	/* Start the game! */
	game.status = GAME_RUNNING;
}

void GAME_move_snake(int x, int y)
{
	if (abs(x) > abs(y))
	{
		if (x < 0) {
			if (game.snake.direction != LEFT)
			game.snake.direction = RIGHT;
			} else {
			if (game.snake.direction != RIGHT)
			game.snake.direction = LEFT;
		}

		} else if (abs(x) < abs(y)) {
		if (y < 0) {
			if (game.snake.direction != UP)
			game.snake.direction = DOWN;
			} else {
			if (game.snake.direction != DOWN)
			game.snake.direction = UP;
		}
	} else
		return;
	
}

void GAME_loop(void)
{
	/* Check if the player lost or won. */

	if (game.status == GAME_LOST) {
		/* Show message. */

		ST7735R_DrawText(ST7735R_WIDTH / 2 - 25, ST7735R_HEIGHT / 2 - 20,
		"YOU LOST!", 255, 255, 255, 0, 0, 0);

		GAME_make_sound(MORSE_LUNG, 0);
		_delay_ms(3000);
		GAME_init();
		
		} else if (game.status == GAME_WON) {
		/* Show message. */

		ST7735R_DrawText(ST7735R_WIDTH / 2 - 40, ST7735R_HEIGHT / 2 - 20,
		"NICE, YOU WON!", 255, 255, 255, 0, 0, 0);
		
		GAME_make_sound(0, 1);
		_delay_ms(4000);
		GAME_init();
	}

	GAME_physics();
	GAME_draw_snake();
	GAME_draw_score();
}

int GAME_boost()
{
	return game.snake.boost;
}

int GAME_buzz_delay(void)
{
	return buzz_delay;
}

/****************************************************************************
* Private Functions Prototypes
****************************************************************************/

static void GAME_init_segment(uint8_t head, int pos_x, int pos_y, int index) {
	
	game.snake.segments[index].color.r = SNAKE_BODY_COLOR_R;
	game.snake.segments[index].color.g = SNAKE_BODY_COLOR_G;
	game.snake.segments[index].color.b = SNAKE_BODY_COLOR_B;

	game.snake.segments[index].head = head;
	game.snake.segments[index].pos.x = pos_x;
	game.snake.segments[index].pos.y = pos_y;
}

static void GAME_draw_snake(void)
{
	/* Remove snake's last segment. */
	GAME_draw_segment(game.snake.tail.pos.x, game.snake.tail.pos.y,
	GAME_WINDOW_BG_R, GAME_WINDOW_BG_G, GAME_WINDOW_BG_B);

	/* Draw first 2 segments. First is new head's location,
	second to cover old head's position. */
	for (int i = 0; i < game.snake.length - 1; i++) {
		if (game.status == GAME_RUNNING && i >= 2) {
			break;
		}
		
		int pos_x = game.snake.segments[i].pos.x;
		int pos_y = game.snake.segments[i].pos.y;
		uint8_t r = game.snake.segments[i].color.r;
		uint8_t g = game.snake.segments[i].color.g;
		uint8_t b = game.snake.segments[i].color.b;
		
		GAME_draw_segment(pos_x, pos_y, r, g, b);

		if (i == 0)
			ST7735R_FilledCircle(pos_x + 3, pos_y + 3 + SCORE_OFFSET, 1,
			GAME_WINDOW_BG_R,
			GAME_WINDOW_BG_G,
			GAME_WINDOW_BG_B);
	}
}

static void GAME_draw_segment(int x, int y, uint8_t r, uint8_t g, uint8_t b)
{
	uint8_t segment_x = 0;
	uint8_t segment_y = 0;
	
	if (x + SEGMENT_WIDTH > GAME_WINDOW_WIDTH) {
		segment_x = 1;
	}
	
	if (y + SEGMENT_WIDTH > GAME_WINDOW_HEIGHT - 1) {
		segment_y = 1;
	}
	
	if (segment_x == 0 && segment_y == 0) {
		ST7735R_FillRect(x, y + SCORE_OFFSET, x + SEGMENT_WIDTH, y + SEGMENT_WIDTH + SCORE_OFFSET, r, g, b);
	} else if (segment_y == 0) {
		ST7735R_FillRect(x, y + SCORE_OFFSET, GAME_WINDOW_WIDTH, y + SEGMENT_WIDTH + SCORE_OFFSET, r, g, b);
	} else
		ST7735R_FillRect(x, y + SCORE_OFFSET, x + SEGMENT_WIDTH, y + SEGMENT_WIDTH + SCORE_OFFSET, r, g, b);

}

static void GAME_physics(void)
{
	int len = game.snake.length - 1;
	buzz_delay = 0;
	/* Check collision */
	int food_collision_index = -1;
	if ((food_collision_index = GAME_check_food_collision()) >= 0)
	{
		game.food[food_collision_index].status = FOOD_EATEN;
		
		if (game.food[food_collision_index].type == BLUE) {
			game.snake.boost = 1;
			game.score += 10;
		} else {
			game.snake.boost = 0;
			game.score += 5;
		}
		GAME_draw_food(food_collision_index);
		GAME_generate_food(food_collision_index);
		GAME_draw_food(food_collision_index);
		GAME_draw_snake();
		GAME_init_segment(0, game.snake.segments[len].pos.x, game.snake.segments[len].pos.y, len + 1);
		
		game.snake.length++;
		buzz_delay = MORSE_SCURT;
	}
	
	if (GAME_check_snake_collsion() == 1)
		game.status = GAME_LOST;
	else if (game.snake.length == 20)
		game.status = GAME_WON;
	else {
		len = game.snake.length - 1;
		game.snake.tail.pos.x = game.snake.segments[len].pos.x;
		game.snake.tail.pos.y = game.snake.segments[len].pos.y;

		for (int i = len; i > 0; i--) {
			game.snake.segments[i].pos.x = game.snake.segments[i - 1].pos.x;
			game.snake.segments[i].pos.y = game.snake.segments[i - 1].pos.y;
		}
		
		int pos_x = game.snake.segments[0].pos.x;
		int pos_y = game.snake.segments[0].pos.y;
		uint8_t direction = game.snake.direction;
		
		if (direction == RIGHT) {
			game.snake.segments[0].pos.x = ( pos_x + SEGMENT_WIDTH + SEGMENT_SPACE + GAME_WINDOW_WIDTH)
			% GAME_WINDOW_WIDTH;
			} else if (direction == LEFT) {
			game.snake.segments[0].pos.x = ( pos_x - SEGMENT_WIDTH - SEGMENT_SPACE + GAME_WINDOW_WIDTH)
			% GAME_WINDOW_WIDTH;
			} else if (direction == DOWN) {
			game.snake.segments[0].pos.y = ( pos_y - SEGMENT_WIDTH - SEGMENT_SPACE + GAME_WINDOW_HEIGHT)
			% GAME_WINDOW_HEIGHT;
			} else {
			game.snake.segments[0].pos.y = ( pos_y + SEGMENT_WIDTH + SEGMENT_SPACE + GAME_WINDOW_HEIGHT)
			% GAME_WINDOW_HEIGHT;
		}
	}
	if (buzz_delay != 0)
		GAME_make_sound(MORSE_SCURT, 0);
}

static int get_rand() {
	rand_state = (rand_state * 1103515245 + 12345) & 0x7fffffff;
	return rand_state;
}

static void GAME_draw_all_food(void)
{
	for (int i = 0; i < FOOD_NUMBER; i++)
		GAME_draw_food(i);
}

static void GAME_generate_food(int index)
{
	int i;
	int pos_x;
	int pos_y;
	int ok;
	int once = 0;

	do {
		ok = 1;
		pos_x = abs(get_rand()) % (GAME_WINDOW_WIDTH - 2 * FOOD_RADIUS) + FOOD_RADIUS;
		pos_y = abs(get_rand()) % (GAME_WINDOW_HEIGHT - 2 * FOOD_RADIUS) + FOOD_RADIUS;
		int food_x = pos_x - FOOD_RADIUS;
		int food_y = pos_y - FOOD_RADIUS;
		int width = 2 * FOOD_RADIUS;
		
		if (once == 0) {
			segment_t snake = game.snake.tail;
		
			if ((snake.pos.x + SEGMENT_WIDTH) >= food_x
			&& snake.pos.x <= (food_x + width)
			&& (snake.pos.y + SEGMENT_WIDTH) >= food_y
			&& snake.pos.y <= (food_y + width))
				ok = 0;
			once++;
		}
		
		for (i = 0; i < game.snake.length - 1; i++) {

			segment_t snake = game.snake.segments[i];
			
			if ((snake.pos.x + SEGMENT_WIDTH) >= food_x
			&& snake.pos.x <= (food_x + width)
			&& (snake.pos.y + SEGMENT_WIDTH) >= food_y
			&& snake.pos.y <= (food_y + width))
			ok = 0;
		}
		
		for (i = 0; i < current_food_items; i++) {
			food_t *other_food = &game.food[i];
			int other_food_x = other_food->pos.x - FOOD_RADIUS;
			int other_food_y = other_food->pos.y - FOOD_RADIUS;
			
			if ((other_food_x + width) >= food_x
			&& other_food_x <= (food_x + width)
			&& (other_food_y + width) >= food_y
			&& other_food_y <= (food_y + width))
			ok = 0;
		}
	} while (ok == 0);
	
	current_food_items++;
	food_t *food = &game.food[index];
	
	if (index % 2) {
		food->type = BLUE;
		food->color.r = FOOD_BLUE_COLOR_R;
		food->color.g = FOOD_BLUE_COLOR_G;
		food->color.b = FOOD_BLUE_COLOR_B;
	} else {
		food->type = RED;
		food->color.r = FOOD_RED_COLOR_R;
		food->color.g = FOOD_RED_COLOR_G;
		food->color.b = FOOD_RED_COLOR_B;
	}
	
	food->pos.x = pos_x;
	food->pos.y = pos_y;
	
	food->status = FOOD_FRESH;
}

static void GAME_init_food(void)
{
	for (int i = 0; i < FOOD_NUMBER; i++)
	GAME_generate_food(i);
}

static void GAME_draw_food(int index)
{
	food_t *food = &game.food[index];
	if (food->status == FOOD_EATEN)
		ST7735R_FilledCircle(food->pos.x, food->pos.y + SCORE_OFFSET, FOOD_RADIUS,
		GAME_WINDOW_BG_R,
		GAME_WINDOW_BG_G,
		GAME_WINDOW_BG_B);
	else
		ST7735R_FilledCircle(food->pos.x, food->pos.y + SCORE_OFFSET, FOOD_RADIUS,
		food->color.r,
		food->color.g,
		food->color.b);
}

static int GAME_check_food_collision(void)
{
	segment_t head = game.snake.segments[0];
	
	/* Consider our food circle a box with width 2 * radius */

	for (int i = 0; i < FOOD_NUMBER; i++) {
		food_t *food = &game.food[i];
		int food_x = food->pos.x - FOOD_RADIUS;
		int food_y = food->pos.y - FOOD_RADIUS;
		int width = 2 * FOOD_RADIUS;
		
		if ((head.pos.x + SEGMENT_WIDTH) >= food_x
		&& head.pos.x <= (food_x + width)
		&& (head.pos.y + SEGMENT_WIDTH) >= food_y
		&& head.pos.y <= (food_y + width))
		return i;
	}
	return -1;
}

static int GAME_check_snake_collsion(void)
{
	segment_t head = game.snake.segments[0];

	for (int i = 1; i < game.snake.length - 1; i++) {

		segment_t snake = game.snake.segments[i];
		
		if ((snake.pos.x + SEGMENT_WIDTH) >= head.pos.x
		&& snake.pos.x <= (head.pos.x + SEGMENT_WIDTH)
		&& (snake.pos.y + SEGMENT_WIDTH) >= head.pos.y
		&& snake.pos.y <= (head.pos.y + SEGMENT_WIDTH))
		return 1;
	}
	
	return 0;
}

static void GAME_draw_score(void)
{
	char buf[30];

	sprintf(buf, "SCORE: %d", game.score);
	ST7735R_DrawText(5, 6, buf, 255, 255, 255, 0, 0, 0);

}

 static void GAME_buzzer_morse(int tip_morse) {
	 int i;
	 for (i = 0; i < tip_morse; ++i) {
		 PORTD ^= (1 << PD0);

		 _delay_ms(1);
	 }
 }

 static void GAME_make_sound(int tip_morse, int type){
	 //Buzzer
	 if (type == 1)
		 for (int i = 0; i < 8; i++) {
			 int note_duration = 1000 / win_sound[i];
			 GAME_buzzer_morse(note_duration);
			 _delay_ms(4);
		 }
	else 
		GAME_buzzer_morse(tip_morse);
	PORTD &= ~(1 << PD0);
}